#!/usr/bin/env python

from datetime import datetime, timedelta
import json
import logging
import sys
import pytz
import pytest
from pycarwings3 import CarwingsError
from pycarwings3.responses import (
    CarwingsLatestBatteryStatusResponse,
    CarwingsLatestClimateControlStatusResponse
)

pytest_plugins = ('pytest_asyncio',)

logging.basicConfig(stream=sys.stdout, level=logging.ERROR)


def test_get_latest_hvac_status_on():
    # Source: original pycarwings2 sources
    climateresponse = """
        {
            "status":200,
            "message":"success",
            "RemoteACRecords":{
                "OperationResult":"START_BATTERY",
                "OperationDateAndTime":"Feb 10, 2016 10:22 PM",
                "RemoteACOperation":"START",
                "ACStartStopDateAndTime":"2018/04/08 10:00",
                "CruisingRangeAcOn":"107712.0",
                "CruisingRangeAcOff":"109344.0",
                "ACStartStopURL":"",
                "PluginState":"NOT_CONNECTED",
                "ACDurationBatterySec":"900",
                "ACDurationPluggedSec":"7200"
            },
            "OperationDateAndTime":""
        }
"""
    response = json.loads(climateresponse)
    status = CarwingsLatestClimateControlStatusResponse(response)
    assert status.is_hvac_running
    assert status.cruising_range_ac_on_km == 107.712
    assert status.cruising_range_ac_off_km == 109.344
    assert status.ac_duration == timedelta(seconds=900)
    assert status.ac_start_stop_date_and_time == datetime(2018, 4, 8, 10, 0, 0, tzinfo=pytz.utc)
    assert status.is_plugged_in is False

def test_get_latest_hvac_status_off():
    # Source: From 24kWh Leaf (Climate is OFF)
    climateresponse = """
{
  "status":200,
  "RemoteACRecords":{
    "OperationResult":"START_BATTERY",
    "OperationDateAndTime":"11-Jan-2019 17:51",
    "RemoteACOperation":"STOP",
    "ACStartStopDateAndTime":"2019/01/11 16:52",
    "CruisingRangeAcOn":"74328.0",
    "CruisingRangeAcOff":"79544.0",
    "ACStartStopURL":"",
    "PluginState":"NOT_CONNECTED",
    "ACDurationBatterySec":"900",
    "ACDurationPluggedSec":"7200"}
}
"""
    response = json.loads(climateresponse)
    status = CarwingsLatestClimateControlStatusResponse(response)
    assert not status.is_hvac_running


def test_get_latest_hvac_status_error():
    # Source: original pycarwings2 sources
    climateresponse = """
        {
            "status":200,
            "RemoteACRecords":{
                "OperationResult":"ELECTRIC_WAVE_ABNORMAL",
                "OperationDateAndTime":"2018/04/08 10:00",
                "RemoteACOperation":"START",
                "ACStartStopDateAndTime":"08-Apr-2018 11:06",
                "ACStartStopURL":"",
                "PluginState":"INVALID",
                "ACDurationBatterySec":"900",
                "ACDurationPluggedSec":"7200",
                "PreAC_unit":"C",
                "PreAC_temp":"22"
            }
        }
"""
    # Assume climate control is off if we get back an empty response
    response = json.loads(climateresponse)
    with pytest.raises(CarwingsError) as excinfo:
        CarwingsLatestClimateControlStatusResponse(response)
        assert 'Should have exception here'

    assert 'could not establish communications with vehicle' in str(
        excinfo.value)


def test_get_latest_hvac_status_empty():
    # Source: 30kWh Leaf as at 13/01/2019
    climateresponse = """
{"status":200,"RemoteACRecords":[]}
"""
    # Assume climate control is off if we get back an empty response
    response = json.loads(climateresponse)
    status = CarwingsLatestClimateControlStatusResponse(response)
    assert not status.is_hvac_running


def test_get_latest_hvac_status_no_ranges_off():
    # Source: unknown leaf in Jan 2019
    # Missing CruisingRangeAcOn and CruisingRangeAcOff
    climateresponse = """
{
  "status": 200,
  "RemoteACRecords": {
    "OperationResult":"START",
    "OperationDateAndTime":"2019/01/24 09:43",
    "RemoteACOperation":"STOP",
    "ACStartStopDateAndTime":"2018/04/08 10:00",
    "ACStartStopURL":"",
    "PluginState":"NOT_CONNECTED",
    "ACDurationBatterySec":"900",
    "ACDurationPluggedSec ":"7200",
    "PreAC_unit":"C",
    "PreAC_temp":"22",
    "Inc_temp ":"12"
  }
}
"""
    # Climate control should be off
    response = json.loads(climateresponse)
    status = CarwingsLatestClimateControlStatusResponse(response)
    assert not status.is_hvac_running


def test_get_latest_hvac_status_no_ranges_on():
    # Source: unknown leaf in Jan 2019
    # Missing CruisingRangeAcOn and CruisingRangeAcOff
    climateresponse = """
{
  "status": 200,
  "RemoteACRecords": {
    "OperationResult":"START",
    "OperationDateAndTime":"2019/01/24 09:43",
    "RemoteACOperation":"START",
    "ACStartStopDateAndTime":"2019-jan-24 10:44",
    "ACStartStopURL":"",
    "PluginState":"NOT_CONNECTED",
    "ACDurationBatterySec":"900",
    "ACDurationPluggedSec ":"7200",
    "PreAC_unit":"C",
    "PreAC_temp":"22",
    "Inc_temp ":"12"
  }
}
"""
    # Assume climate control should be on
    response = json.loads(climateresponse)
    status = CarwingsLatestClimateControlStatusResponse(response)
    assert status.is_hvac_running


def test_get_latest_battery_status_24kWh_30percent_0degredation():
    # not connected to a charger
    batteryresponse = """
        {
            "status":200,
            "message":"success",
            "BatteryStatusRecords":{
                "OperationResult":"START",
                "OperationDateAndTime":"Feb  9, 2016 11:09 PM",
                "BatteryStatus":{
                    "BatteryChargingStatus":"NOT_CHARGING",
                    "BatteryCapacity":"12",
                    "BatteryRemainingAmount":"3",
                    "BatteryRemainingAmountWH":"",
                    "BatteryRemainingAmountkWH":""
                },
                "PluginState":"NOT_CONNECTED",
                "CruisingRangeAcOn":"39192.0",
                "CruisingRangeAcOff":"39744.0",
                "TimeRequiredToFull":{
                    "HourRequiredToFull":"18",
                    "MinutesRequiredToFull":"30"
                },
                "TimeRequiredToFull200":{
                    "HourRequiredToFull":"6",
                    "MinutesRequiredToFull":"0"
                },
                "TimeRequiredToFull200_6kW":{
                    "HourRequiredToFull":"4",
                    "MinutesRequiredToFull":"0"
                },
                "NotificationDateAndTime":"2016/02/10 04:10",
                "TargetDate":"2016/02/10 04:09"
            }
        }
"""
    response = json.loads(batteryresponse)
    status = CarwingsLatestBatteryStatusResponse(response)
    assert status.battery_percent == 100 * 3 / 12


def test_get_latest_battery_status_24kWh_50percent_1degredation():
    # Charging, but with 50% charged, with 11 bars max charge (1 bar lost)
    batteryresponse = """
{
  "status": 200,
  "BatteryStatusRecords": {
    "OperationResult": "START",
    "OperationDateAndTime": "26-Jan-2019 22:22",
    "BatteryStatus": {
      "BatteryChargingStatus": "NORMAL_CHARGING",
      "BatteryCapacity": "11",
      "BatteryRemainingAmount": "6",
      "BatteryRemainingAmountWH": "",
      "BatteryRemainingAmountkWH": ""
    },
    "PluginState": "CONNECTED",
    "CruisingRangeAcOn": "66024.0",
    "CruisingRangeAcOff": "69168.0",
    "TimeRequiredToFull200_6kW": {
      "HourRequiredToFull": "3",
      "MinutesRequiredToFull": "0"
    },
    "NotificationDateAndTime": "2019/01/26 21:22",
    "TargetDate": "2019/01/26 21:22"
  }
}
"""
    response = json.loads(batteryresponse)
    status = CarwingsLatestBatteryStatusResponse(response)
    assert status.battery_percent == 100 * 6 / 12


def test_get_latest_battery_status_24kWh_91percent_1degredation():
    # Not charging, but with 11 bars charged, 11 bars max charge (1 bar lost)
    batteryresponse = """
{
  "status": 200,
  "BatteryStatusRecords": {
    "OperationResult": "START",
    "OperationDateAndTime": "09-Jan-2019 20:45",
    "BatteryStatus": {
      "BatteryChargingStatus": "NOT_CHARGING",
      "BatteryCapacity": "11",
      "BatteryRemainingAmount": "11",
      "BatteryRemainingAmountWH": "",
      "BatteryRemainingAmountkWH": ""
    },
    "PluginState": "NOT_CONNECTED",
    "CruisingRangeAcOn": "97704.0",
    "CruisingRangeAcOff": "105984.0",
    "TimeRequiredToFull": {
      "HourRequiredToFull": "3",
      "MinutesRequiredToFull": "30"
    },
    "TimeRequiredToFull200": {
      "HourRequiredToFull": "2",
      "MinutesRequiredToFull": "30"
    },
    "TimeRequiredToFull200_6kW": {
      "HourRequiredToFull": "2",
      "MinutesRequiredToFull": "0"
    },
    "NotificationDateAndTime": "2019/01/09 19:45",
    "TargetDate": "2019/01/09 19:45"
  }
}
"""
    response = json.loads(batteryresponse)
    status = CarwingsLatestBatteryStatusResponse(response)
    assert status.battery_percent == 100 * 11 / 12


def test_get_latest_battery_status_24kWh_100percent_1degredation():
    # Not charging, but fully charged, 1 bar lost
    batteryresponse = """
{
  "status":200,
  "BatteryStatusRecords":{
    "OperationResult":"START",
    "OperationDateAndTime":"03-Feb-2019 23:50",
    "BatteryStatus":{
      "BatteryChargingStatus":"NOT_CHARGING",
      "BatteryCapacity":"11",
      "BatteryRemainingAmount":"12",
      "BatteryRemainingAmountWH":"",
      "BatteryRemainingAmountkWH":""
    },
    "PluginState":"NOT_CONNECTED",
    "CruisingRangeAcOn":"116112.0",
    "CruisingRangeAcOff":"131856.0",
    "TimeRequiredToFull":{
      "HourRequiredToFull":"0",
      "MinutesRequiredToFull":"40"
    },
    "TimeRequiredToFull200":{
      "HourRequiredToFull":"0",
      "MinutesRequiredToFull":"40"
    },
    "TimeRequiredToFull200_6kW":{
      "HourRequiredToFull":"0",
      "MinutesRequiredToFull":"40"
    },
    "NotificationDateAndTime":"2019/02/03 22:50",
    "TargetDate":"2019/02/03 22:50"
  }
}
"""
    response = json.loads(batteryresponse)
    status = CarwingsLatestBatteryStatusResponse(response)
    assert status.battery_percent == 100    # = 100 * 12 / 12


def test_get_latest_battery_status_30kWh_91percent_0degredation():
    # not connected to a charger - as at 21/01/2019 20:01 (for a 30kWh leaf)
    batteryresponse = """
        {
            "status":200,
            "BatteryStatusRecords": {
                "OperationResult":"START",
                "OperationDateAndTime":"21-Jan-2019 13:29",
                "BatteryStatus":{
                    "BatteryChargingStatus":"NOT_CHARGING",
                    "BatteryCapacity":"240",
                    "BatteryRemainingAmount":"220",
                    "BatteryRemainingAmountWH":"24480",
                    "BatteryRemainingAmountkWH":"",
                    "SOC":{
                        "Value":"91"
                    }
                },
                "PluginState":"NOT_CONNECTED",
                "CruisingRangeAcOn":"146000",
                "CruisingRangeAcOff":"168000",
                "TimeRequiredToFull":{
                    "HourRequiredToFull":"4",
                    "MinutesRequiredToFull":"30"
                },
                "TimeRequiredToFull200":{
                    "HourRequiredToFull":"3"
                    ,"MinutesRequiredToFull":"0"
                },
                "TimeRequiredToFull200_6kW":{
                    "HourRequiredToFull":"1",
                    "MinutesRequiredToFull":"30"
                },
                "NotificationDateAndTime":"2019/01/21 13:29",
                "TargetDate":"2019/01/21 13:29"
            }
        }
"""
    response = json.loads(batteryresponse)
    status = CarwingsLatestBatteryStatusResponse(response)
    assert status.battery_percent == 91



# Copyright 2016 Jason Horne
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import timedelta, datetime
import pytz

import pycarwings3

log = logging.getLogger(__name__)


def _time_remaining(t):
    minutes = float(0)
    if t:
        if ("hours" in t) and t["hours"]:
            minutes = 60 * float(t["hours"])
        elif ("HourRequiredToFull" in t) and t["HourRequiredToFull"]:
            minutes = 60 * float(t["HourRequiredToFull"])
        if ("minutes" in t) and t["minutes"]:
            minutes += float(t["minutes"])
        elif ("MinutesRequiredToFull" in t) and t["MinutesRequiredToFull"]:
            minutes += float(t["MinutesRequiredToFull"])

    return minutes


class CarwingsResponse:
    def __init__(self, response):
        op_result = None
        if ("operationResult" in response):
            op_result = response["operationResult"]
        elif ("OperationResult" in response):
            op_result = response["OperationResult"]

        # seems to indicate that the vehicle cannot be reached
        if ("ELECTRIC_WAVE_ABNORMAL" == op_result):
            log.error("could not establish communications with vehicle")
            raise pycarwings3.CarwingsError("could not establish communications with vehicle")

    def _set_cruising_ranges(self, status, off_key="cruisingRangeAcOff", on_key="cruisingRangeAcOn"):
        if off_key in status:
            self.cruising_range_ac_off_km = float(status[off_key]) / 1000
        if on_key in status:
            self.cruising_range_ac_on_km = float(status[on_key]) / 1000

    def _set_timestamp(self, status):
        self.timestamp = datetime.strptime(status["timeStamp"], "%Y-%m-%d %H:%M:%S")   # "2016-01-02 17:17:38"


class CarwingsInitialAppResponse(CarwingsResponse):
    def __init__(self, response):
        CarwingsResponse.__init__(self, response)
        self.baseprm = response["baseprm"]


class CarwingsLoginResponse(CarwingsResponse):
    """
        example JSON response to login:
        {
            "status":200,
            "message":"success",
            "sessionId":"12345678-1234-1234-1234-1234567890",
            "VehicleInfoList": {
                "VehicleInfo": [
                    {
                        "charger20066":"false",
                        "nickname":"LEAF",
                        "telematicsEnabled":"true",
                        "vin":"1ABCDEFG2HIJKLM3N"
                    }
                ],
                "vehicleInfo": [
                    {
                        "charger20066":"false",
                        "nickname":"LEAF",
                        "telematicsEnabled":"true",
                        "vin":"1ABCDEFG2HIJKLM3N"
                    }
                ]
            },
            "vehicle": {
                "profile": {
                    "vin":"1ABCDEFG2HIJKLM3N",
                    "gdcUserId":"FG12345678",
                    "gdcPassword":"password",
                    "encAuthToken":"ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ",
                    "dcmId":"123456789012",
                    "nickname":"Alpha124",
                    "status":"ACCEPTED",
                    "statusDate": "Aug 15, 2015 07:00 PM"
                }
            },
            "EncAuthToken":"ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ",
            "CustomerInfo": {
                "UserId":"AB12345678",
                "Language":"en-US",
                "Timezone":"America/New_York",
                "RegionCode":"NNA",
                "OwnerId":"1234567890",
                "Nickname":"Bravo456",
                "Country":"US",
                "VehicleImage":"/content/language/default/images/img/ph_car.jpg",
                "UserVehicleBoundDurationSec":"999971200",
                "VehicleInfo": {
                    "VIN":"1ABCDEFG2HIJKLM3N",
                    "DCMID":"201212345678",
                    "SIMID":"12345678901234567890",
                    "NAVIID":"1234567890",
                    "EncryptedNAVIID":"1234567890ABCDEFGHIJKLMNOP",
                    "MSN":"123456789012345",
                    "LastVehicleLoginTime":"",
                    "UserVehicleBoundTime":"2015-08-17T14:16:32Z",
                    "LastDCMUseTime":""
                }
            },
            "UserInfoRevisionNo":"1"
        }

        example JSON response to login when subscription has expired. Resubscribe via the Nissan Connect EV website.
        {
            "status":"AAS-604",
            "message":"INVALID PARAMS"
        }
    """
    def __init__(self, response):
        CarwingsResponse.__init__(self, response)

        profile = response["vehicle"]["profile"]
        self.gdc_user_id = profile["gdcUserId"]
        self.dcm_id = profile["dcmId"]
        self.vin = profile["vin"]

        # vehicleInfo block may be top level, or contained in a VehicleInfoList object;
        # why it's sometimes one way and sometimes another is not clear.
        if "VehicleInfoList" in response:
            self.nickname = response["VehicleInfoList"]["vehicleInfo"][0]["nickname"]
            self.custom_sessionid = response["VehicleInfoList"]["vehicleInfo"][0]["custom_sessionid"]
        elif "vehicleInfo" in response:
            self.nickname = response["vehicleInfo"][0]["nickname"]
            self.custom_sessionid = response["vehicleInfo"][0]["custom_sessionid"]

        customer_info = response["CustomerInfo"]
        self.tz = customer_info["Timezone"]
        self.language = customer_info["Language"]
        self.user_vehicle_bound_time = \
            customer_info["VehicleInfo"].get("UserVehicleBoundTime",
                                             "1970-01-01T00:00:00Z")

        self.leafs = [{
            "vin": self.vin,
            "nickname": self.nickname,
            "bound_time": self.user_vehicle_bound_time
        }]


class CarwingsBatteryStatusResponse(CarwingsResponse):
    """
        Note that before December 2018 this returned a response. Between Dec-2018 and Aug-2019
        it did not return a response from the Nissan Servers.  As of Aug 2019 a response is now
        returned again.

        # Original
        {
            "status": 200,
            "message": "success",
            "responseFlag": "1",
            "operationResult": "START",
            "timeStamp": "2016-01-02 17:17:38",
            "cruisingRangeAcOn": "115328.0",
            "cruisingRangeAcOff": "117024.0",
            "currentChargeLevel": "0",
            "chargeMode": "220V",
            "pluginState": "CONNECTED",
            "charging": "YES",
            "chargeStatus": "CT",
            "batteryDegradation": "10",
            "batteryCapacity": "12",
            "timeRequiredToFull": {
                "hours": "",
                "minutes": ""
            },
            "timeRequiredToFull200": {
                "hours": "",
                "minutes": ""
            },
            "timeRequiredToFull200_6kW": {
                "hours": "",
                "minutes": ""
            }
        }

        # As at 21/01/2019 for a 30kWh Leaf now seems that
        # BatteryStatusCheckResultRequest.php always returns this
        # regardless of battery status.
        {
            "status":200,
            "responseFlag":"0"
        }

        {
            "status":200,
            "message":"success",
            "responseFlag":"1",
            "operationResult":"START",
            "timeStamp":"2016-02-14 20:28:45",
            "cruisingRangeAcOn":"107136.0",
            "cruisingRangeAcOff":"115776.0",
            "currentChargeLevel":"0",
            "chargeMode":"NOT_CHARGING",
            "pluginState":"QC_CONNECTED",
            "charging":"YES",
            "chargeStatus":"CT",
            "batteryDegradation":"11",
            "batteryCapacity":"12",
            "timeRequiredToFull":{
                "hours":"",
                "minutes":""
            },
            "timeRequiredToFull200":{
                "hours":"",
                "minutes":""
            },
            "timeRequiredToFull200_6kW":{
                "hours":"",
                "minutes":""
            }
        }

        # As at 22/08/2019 for a 30kWh Leaf now seesm that
        # BatteryStatusCheckResultRequest.php returns data again
        # after polling a number of times.
        {
                "status": 200,
                "responseFlag": "1",
                "operationResult": "START",
                "timeStamp": "2019-08-22 10:26:51",
                "cruisingRangeAcOn": "129000.0",
                "cruisingRangeAcOff": "132000.0",
                "currentChargeLevel": "0",
                "chargeMode": "NOT_CHARGING",
                "pluginState": "NOT_CONNECTED",
                "charging": "NO",
                "chargeStatus": "0",
                "batteryDegradation": "180",
                "batteryCapacity": "240",
                "timeRequiredToFull": {
                        "hours": "11",
                        "minutes": "30"
                },
                "timeRequiredToFull200": {
                        "hours": "6",
                        "minutes": "30"
                },
                "timeRequiredToFull200_6kW": {
                        "hours": "2",
                        "minutes": "30"
                }
        }

    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        self._set_timestamp(status)
        self._set_cruising_ranges(status)

        self.answer = status

        self.battery_capacity = status["batteryCapacity"]
        self.battery_degradation = status["batteryDegradation"]

        self.is_connected = ("NOT_CONNECTED" != status["pluginState"])    # fun double negative
        self.plugin_state = status["pluginState"]

        self.charging_status = status["chargeMode"]

        self.is_charging = ("YES" == status["charging"])

        self.is_quick_charging = ("RAPIDLY_CHARGING" == status["chargeMode"])
        self.is_connected_to_quick_charger = ("QC_CONNECTED" == status["pluginState"])

        self.time_to_full_trickle = timedelta(minutes=_time_remaining(status["timeRequiredToFull"]))
        self.time_to_full_l2 = timedelta(minutes=_time_remaining(status["timeRequiredToFull200"]))
        self.time_to_full_l2_6kw = timedelta(minutes=_time_remaining(status["timeRequiredToFull200_6kW"]))

        # For some leafs the battery_percent is not returned
        self.battery_percent = 100 * float(self.battery_degradation) / 12


class CarwingsLatestClimateControlStatusResponse(CarwingsResponse):
    """
    climate control on:
        {
            "status":200,
            "message":"success",
            "RemoteACRecords":{
                "OperationResult":"START_BATTERY",
                "OperationDateAndTime":"Feb 10, 2016 10:22 PM",
                "RemoteACOperation":"START",
                "ACStartStopDateAndTime":"2018/04/08 10:00",
                "CruisingRangeAcOn":"107712.0",
                "CruisingRangeAcOff":"109344.0",
                "ACStartStopURL":"",
                "PluginState":"NOT_CONNECTED",
                "ACDurationBatterySec":"900",
                "ACDurationPluggedSec":"7200"
            },
            "OperationDateAndTime":""
        }

    climate control off:
        {
            "status":200,
            "message":"success",
            "RemoteACRecords":{
                "OperationResult":"START",
                "OperationDateAndTime":"Feb 10, 2016 10:26 PM",
                "RemoteACOperation":"STOP",
                "ACStartStopDateAndTime":"2018/04/08 10:00",
                "CruisingRangeAcOn":"111936.0",
                "CruisingRangeAcOff":"113632.0",
                "ACStartStopURL":"",
                "PluginState":"NOT_CONNECTED",
                "ACDurationBatterySec":"900",
                "ACDurationPluggedSec":"7200"
            },
            "OperationDateAndTime":""
        }

    error:
        {
            "status":200,
            "RemoteACRecords":{
                "OperationResult":"ELECTRIC_WAVE_ABNORMAL",
                "OperationDateAndTime":"2018/04/08 10:00",
                "RemoteACOperation":"START",
                "ACStartStopDateAndTime":"2018/04/08 10:00",
                "ACStartStopURL":"",
                "PluginState":"INVALID",
                "ACDurationBatterySec":"900",
                "ACDurationPluggedSec":"7200",
                "PreAC_unit":"C",
                "PreAC_temp":"22"
            }
        }
    noinfo (from a 2014 24kWh Leaf):
        {
            "status":200,
            "RemoteACRecords":[]
        }

    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status["RemoteACRecords"])
        racr = status["RemoteACRecords"]

        self._set_cruising_ranges(racr, on_key="CruisingRangeAcOn", off_key="CruisingRangeAcOff")

        if isinstance(racr, dict):
            self._operation_date_and_time = racr["OperationDateAndTime"]

        # If no empty RemoteACRecords list is returned then assume CC is off.
        if not isinstance(racr, dict):
            self.is_hvac_running = False
        else:
            # Seems to be running only if both of these contain "START".
            self.is_hvac_running = (
                racr["OperationResult"] and
                racr["OperationResult"].startswith("START") and
                racr["RemoteACOperation"] == "START"
            )

        if isinstance(racr, dict):

            try:
                self.is_plugged_in = ("NOT_CONNECTED" != racr["PluginState"])
            except KeyError:
                self.is_plugged_in = None

            try:
                self.ac_duration = timedelta(seconds=int(racr["ACDurationPluggedSec"] if self.is_plugged_in else racr["ACDurationBatterySec"]))
            except KeyError:
                self.ac_duration = None

            try:
                # "2018/04/08 10:00"
                self.ac_start_stop_date_and_time = datetime.strptime(racr["ACStartStopDateAndTime"], "%Y/%m/%d %H:%M").replace(tzinfo=pytz.utc)
            except (KeyError, ValueError):
                self.ac_start_stop_date_and_time = None
        else:
            self.is_plugged_in = None
            self.ac_duration = None
            self.ac_start_stop_date_and_time = None

        # "Feb 10, 2016 10:26 PM"
        try:
            if isinstance(racr, dict):
                self.timestamp = datetime.strptime(racr["OperationDateAndTime"],
                                                    "%b %d, %Y %I:%M %p").replace(tzinfo=pytz.utc)
            else:
                self.timestamp = None
        except (KeyError, ValueError):
            self.timestamp = None

    def __eq__(self, other):
        if not isinstance(other, CarwingsLatestClimateControlStatusResponse):
            return NotImplemented

        if self._operation_date_and_time is None:
            return NotImplemented

        return (
            self._operation_date_and_time == other._operation_date_and_time
        )


class CarwingsStartClimateControlResponse(CarwingsResponse):
    """
        {
            "status":200,
            "message":"success",
            "responseFlag":"1",
            "operationResult":"START_BATTERY",
            "acContinueTime":"15",
            "cruisingRangeAcOn":"106400.0",
            "cruisingRangeAcOff":"107920.0",
            "timeStamp":"2016-02-05 12:59:46",
            "hvacStatus":"ON"
        }
    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        self._set_timestamp(status)
        self._set_cruising_ranges(status)

        self.operation_result = status["operationResult"]     # e.g. "START_BATTERY", ...?
        self.ac_continue_time = timedelta(minutes=float(status["acContinueTime"]))
        self.hvac_status = status["hvacStatus"]               # "ON" or "OFF"
        self.is_hvac_running = ("ON" == self.hvac_status)


class CarwingsStopClimateControlResponse(CarwingsResponse):
    """
        {
            "status":200,
            "message":"success",
            "responseFlag":"1",
            "operationResult":"START",
            "timeStamp":"2016-02-09 03:32:51",
            "hvacStatus":"OFF"
        }
    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        self._set_timestamp(status)
        self.hvac_status = status["hvacStatus"]                # "ON" or "OFF"
        self.is_hvac_running = ("ON" == self.hvac_status)


class CarwingsClimateControlScheduleResponse(CarwingsResponse):
    """
        {
            "status":200,
            "message":"success",
            "LastScheduledTime":"Feb  9, 2016 05:39 PM",
            "ExecuteTime":"2016-02-10 01:00:00",
            "DisplayExecuteTime":"Feb  9, 2016 08:00 PM",
            "TargetDate":"2016/02/10 01:00"
        }
    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        self.display_execute_time = status["DisplayExecuteTime"]         # displayable, timezone-adjusted
        self.execute_time = datetime.strptime(status["ExecuteTime"] + " UTC", "%Y-%m-%d %H:%M:%S %Z")   # GMT
        self.display_last_scheduled_time = status["LastScheduledTime"]   # displayable, timezone-adjusted
        self.last_scheduled_time = datetime.strptime(status["LastScheduledTime"], "%b %d, %Y %I:%M %p")
        # unknown purpose; don't surface to avoid confusion
        # self.target_date = status["TargetDate"]


class CarwingsDrivingAnalysisResponse(CarwingsResponse):
    """
        {
            "status":200,
            "message":"success",
            "DriveAnalysisBasicScreenResponsePersonalData": {
                "DateSummary":{
                    "TargetDate":"2016-02-03",
                    "ElectricMileage":"4.4",
                    "ElectricMileageLevel":"3",
                    "PowerConsumptMoter":"295.2",
                    "PowerConsumptMoterLevel":"4",
                    "PowerConsumptMinus":"84.8",
                    "PowerConsumptMinusLevel":"3",
                    "PowerConsumptAUX":"17.1",
                    "PowerConsumptAUXLevel":"5",
                    "DisplayDate":"Feb  3, 16"
                },
                "ElectricCostScale":"miles/kWh"
            },
            "AdviceList":{
                "Advice":{
                    "title":"World Number of Trips Rankings (last week):",
                    "body":"The highest number of trips driven was 130 by a driver located in Japan."
                }
            }
        }
    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        summary = status["DriveAnalysisBasicScreenResponsePersonalData"]["DateSummary"]

        # avg energy economy, in units of 'electric_cost_scale' (e.g. miles/kWh)
        self.electric_mileage = summary["ElectricMileage"]
        # rating for above, scale of 1-5
        self.electric_mileage_level = summary["ElectricMileageLevel"]

        # "acceleration performance": "electricity used for motor activation over 1km", Watt-Hours
        self.power_consumption_moter = summary["PowerConsumptMoter"]
        # rating for above, scale of 1-5
        self.power_consumption_moter_level = summary["PowerConsumptMoterLevel"]

        # Watt-Hours generated by braking
        self.power_consumption_minus = summary["PowerConsumptMinus"]
        # rating for above, scale of 1-5
        self.power_consumption_minus_level = summary["PowerConsumptMinusLevel"]

        # Electricity used by aux devices, Watt-Hours
        self.power_consumption_aux = summary["PowerConsumptAUX"]
        # rating for above, scale of 1-5
        self.power_consumption_aux_level = summary["PowerConsumptAUXLevel"]

        self.display_date = summary["DisplayDate"]     # "Feb  3, 16"
        self.target_date = summary["TargetDate"]       # "2016-02-03"

        self.electric_cost_scale = status["DriveAnalysisBasicScreenResponsePersonalData"]["ElectricCostScale"]

        self.advice = [status["AdviceList"]["Advice"]]  # will contain "title" and "body"

    @property
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.target_date == other.target_date
                and self.electric_mileage == other.electric_mileage
                and self.power_consumption_aux == other.power_consumption_aux
                and self.power_consumption_minus == other.power_consumption_minus
                and self.power_consumption_moter == other.power_consumption_moter)


class CarwingsLatestBatteryStatusResponse(CarwingsResponse):
    """
        # not connected to a charger
        {
            "status":200,
            "message":"success",
            "BatteryStatusRecords":{
                "OperationResult":"START",
                "OperationDateAndTime":"Feb  9, 2016 11:09 PM",
                "BatteryStatus":{
                    "BatteryChargingStatus":"NOT_CHARGING",
                    "BatteryCapacity":"12",
                    "BatteryRemainingAmount":"3",
                    "BatteryRemainingAmountWH":"",
                    "BatteryRemainingAmountkWH":""
                },
                "PluginState":"NOT_CONNECTED",
                "CruisingRangeAcOn":"39192.0",
                "CruisingRangeAcOff":"39744.0",
                "TimeRequiredToFull":{              # 120V
                    "HourRequiredToFull":"18",
                    "MinutesRequiredToFull":"30"
                },
                "TimeRequiredToFull200":{           # 240V, 3kW
                    "HourRequiredToFull":"6",
                    "MinutesRequiredToFull":"0"
                },
                "TimeRequiredToFull200_6kW":{       # 240V, 6kW
                    "HourRequiredToFull":"4",
                    "MinutesRequiredToFull":"0"
                },
                "NotificationDateAndTime":"2016/02/10 04:10",
                "TargetDate":"2016/02/10 04:09"
            }
        }

        # not connected to a charger - as at 21/01/2019 20:01 (for a 30kWh leaf)
        {
            "status":200,
            "BatteryStatusRecords": {
                "OperationResult":"START",
                "OperationDateAndTime":"21-Jan-2019 13:29",
                "BatteryStatus":{
                    "BatteryChargingStatus":"NOT_CHARGING",
                    "BatteryCapacity":"240",
                    "BatteryRemainingAmount":"220",
                    "BatteryRemainingAmountWH":"24480",
                    "BatteryRemainingAmountkWH":"",
                    "SOC":{
                        "Value":"91"
                    }
                },
                "PluginState":"NOT_CONNECTED",
                "CruisingRangeAcOn":"146000",
                "CruisingRangeAcOff":"168000",
                "TimeRequiredToFull":{
                    "HourRequiredToFull":"4",
                    "MinutesRequiredToFull":"30"
                },
                "TimeRequiredToFull200":{
                    "HourRequiredToFull":"3"
                    ,"MinutesRequiredToFull":"0"
                },
                "TimeRequiredToFull200_6kW":{
                    "HourRequiredToFull":"1",
                    "MinutesRequiredToFull":"30"
                },
                "NotificationDateAndTime":"2019/01/21 13:29",
                "TargetDate":"2019/01/21 13:29"
            }
        }


        # connected to a quick charger
        {
            "status":200,
            "message":"success",
            "BatteryStatusRecords":{
                "OperationResult":"START",
                "OperationDateAndTime":"Feb 14, 2016 03:28 PM",
                "BatteryStatus":{
                    "BatteryChargingStatus":"RAPIDLY_CHARGING",
                    "BatteryCapacity":"12",
                    "BatteryRemainingAmount":"11",
                    "BatteryRemainingAmountWH":"",
                    "BatteryRemainingAmountkWH":""
                },
                "PluginState":"QC_CONNECTED",
                "CruisingRangeAcOn":"107136.0",
                "CruisingRangeAcOff":"115776.0",
                "NotificationDateAndTime":"2016/02/14 20:28",
                "TargetDate":"2016/02/14 20:28"
            }
        }

        # connected to a charging station
        {
          "status": 200,
          "message": "success",
          "BatteryStatusRecords": {
            "OperationResult": "START",
            "OperationDateAndTime": "Feb 19, 2016 12:12 PM",
            "BatteryStatus": {
              "BatteryChargingStatus": "NORMAL_CHARGING",
              "BatteryCapacity": "12",
              "BatteryRemainingAmount": "12",
              "BatteryRemainingAmountWH": "",
              "BatteryRemainingAmountkWH": ""
            },
            "PluginState": "CONNECTED",
            "CruisingRangeAcOn": "132000.0",
            "CruisingRangeAcOff": "134000.0",
            "TimeRequiredToFull200_6kW": {
              "HourRequiredToFull": "0",
              "MinutesRequiredToFull": "40"
            },
            "NotificationDateAndTime": "2016/02/19 17:12",
            "TargetDate": "2016/02/19 17:12"
          }
        }
    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status["BatteryStatusRecords"])
        self.answer = status

        recs = status["BatteryStatusRecords"]

        bs = recs["BatteryStatus"]
        self.battery_capacity = bs["BatteryCapacity"]
        self.battery_remaining_amount = bs["BatteryRemainingAmount"]
        self.battery_remaining_amount_wh = bs["BatteryRemainingAmountWH"]
        self.charging_status = bs["BatteryChargingStatus"]
        self.is_charging = ("NOT_CHARGING" != bs["BatteryChargingStatus"])    # double negatives are fun
        self.is_quick_charging = ("RAPIDLY_CHARGING" == bs["BatteryChargingStatus"])

        self.plugin_state = recs["PluginState"]
        self.is_connected = ("NOT_CONNECTED" != recs["PluginState"])          # another double negative
        self.is_connected_to_quick_charger = ("QC_CONNECTED" == recs["PluginState"])

        self._set_cruising_ranges(recs, off_key="CruisingRangeAcOff", on_key="CruisingRangeAcOn")

        if "TimeRequiredToFull" in recs:
            self.time_to_full_trickle = timedelta(minutes=_time_remaining(recs["TimeRequiredToFull"]))
        else:
            self.time_to_full_trickle = None

        if "TimeRequiredToFull200" in recs:
            self.time_to_full_l2 = timedelta(minutes=_time_remaining(recs["TimeRequiredToFull200"]))
        else:
            self.time_to_full_l2 = None

        if "TimeRequiredToFull200_6kW" in recs:
            self.time_to_full_l2_6kw = timedelta(minutes=_time_remaining(recs["TimeRequiredToFull200_6kW"]))
        else:
            self.time_to_full_l2_6kw = None

        if float(self.battery_capacity) == 0:
            log.debug("battery_capacity=0, status=%s", status)
            self.battery_percent = 0
        else:
            if self.battery_remaining_amount == '':
                self.battery_percent = 0
            else:
                self.battery_percent = 100 * float(self.battery_remaining_amount) / 12

        # Leaf 2016 has SOC (State Of Charge) in BatteryStatus, a more accurate battery_percentage
        if "SOC" in bs:
            self.state_of_charge = bs["SOC"]["Value"]
            # Update battery_percent with more accurate version
            self.battery_percent = float(self.state_of_charge)
        else:
            self.state_of_charge = None

        # "NotificationDateAndTime":"2016/02/14 20:28",
        self.timestamp = datetime.strptime(
            recs["NotificationDateAndTime"],
            "%Y/%m/%d %H:%M"
        ).replace(tzinfo=pytz.utc)

    def __eq__(self, other):
        if not isinstance(other, CarwingsLatestBatteryStatusResponse):
            return NotImplemented

        return (
            self.timestamp == other.timestamp
        )

class CarwingsElectricRateSimulationResponse(CarwingsResponse):
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        r = status["PriceSimulatorDetailInfoResponsePersonalData"]
        t = r["PriceSimulatorTotalInfo"]

        self.month = r["DisplayMonth"]     # e.g. "Feb/2016"
        self.travellist = r["PriceSimulatorDetailInfoDateList"]["PriceSimulatorDetailInfoDate"]
        self.total_number_of_trips = t["TotalNumberOfTrips"]
        self.total_power_consumption = t["TotalPowerConsumptTotal"]                # in kWh
        self.total_acceleration_power_consumption = t["TotalPowerConsumptMoter"]   # in kWh
        self.total_power_regenerated_in_braking = t["TotalPowerConsumptMinus"]     # in kWh
        self.total_travel_distance_km = float(t["TotalTravelDistance"]) / 1000     # assumed to be in meters

        self.total_electric_mileage = t["TotalElectricMileage"]
        self.total_co2_reduction = t["TotalCO2Reductiont"]   # (yep, extra 't' at the end)

        self.electricity_rate = r["ElectricPrice"]
        self.electric_bill = r["ElectricBill"]
        self.electric_cost_scale = r["ElectricCostScale"]    # e.g. "miles/kWh"


class CarwingsMyCarFinderResponse(CarwingsResponse):
    """
        {
            "Location": {
                "Country": "",
                "Home": "OUTSIDE",
                "LatitudeDeg": "69",
                "LatitudeMin": "41",
                "LatitudeMode": "NORTH",
                "LatitudeSec": "5540",
                "LocationType": "WGS84",
                "LongitudeDeg": "18",
                "LongitudeMin": "38",
                "LongitudeMode": "EAST",
                "LongitudeSec": "2506",
                "Position": "UNAVAILABLE"
            },
            "TargetDate": "2017/11/29 20:02",
            "lat": "69.698722222222",
            "lng": "18.640294444444",
            "receivedDate": "2017/11/29 20:02",
            "responseFlag": "1",
            "resultCode": "1",
            "status": 200,
            "timeStamp": "2017-11-29 20:02:45"
        }
    """
    def __init__(self, status):
        CarwingsResponse.__init__(self, status)

        self.latitude = status["lat"]
        self.longitude = status["lng"]



# Copyright 2016 Jason Horne
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
When logging in, you must specify a geographic 'region' parameter. The only
known values for this are as follows:

    NNA  : USA
    NE   : Europe
    NCI  : Canada
    NMA  : Australia
    NML  : Japan

Information about Nissan on the web (e.g. http://nissannews.com/en-US/nissan/usa/pages/executive-bios)
suggests others (this page suggests NMEX for Mexico, NLAC for Latin America) but
these have not been confirmed.

There are three asynchronous operations in this API, paired with three follow-up
"status check" methods.

    request_update           -> get_status_from_update
    start_climate_control    -> get_start_climate_control_result
    stop_climate_control     -> get_stop_climate_control_result

The asynchronous operations immediately return a 'result key', which
is then supplied as a parameter for the corresponding status check method.

Here's an example response from an asynchronous operation, showing the result key:

    {
        "status":200,
        "userId":"user@domain.com",
        "vin":"1ABCDEFG2HIJKLM3N",
        "resultKey":"12345678901234567890123456789012345678901234567890"
    }

The status check methods return a JSON blob containing a 'responseFlag' property.
If the communications are complete, the response flag value will be the string "1";
otherwise the value will be the string "0". You just gotta poll until you get a
"1" back. Note that the official app seems to poll every 20 seconds.

Example 'no response yet' result from a status check invocation:

    {
        "status":200,
        "responseFlag":"0"
    }

When the responseFlag does come back as "1", there will also be an "operationResult"
property. If there was an error communicating with the vehicle, it seems that
this field will contain the value "ELECTRIC_WAVE_ABNORMAL". Odd.

"""

import json
import logging
from datetime import date, datetime
from .responses import (
    CarwingsInitialAppResponse,
    CarwingsLoginResponse,
    CarwingsBatteryStatusResponse,
    CarwingsStartClimateControlResponse,
    CarwingsStopClimateControlResponse,
    CarwingsDrivingAnalysisResponse,
    CarwingsLatestBatteryStatusResponse,
    CarwingsLatestClimateControlStatusResponse,
    CarwingsElectricRateSimulationResponse,
    CarwingsClimateControlScheduleResponse,
    CarwingsMyCarFinderResponse,
)
import base64
from Crypto.Cipher import Blowfish
from aiohttp import ClientError, ClientSession, ClientTimeout, ContentTypeError
from asyncio import Semaphore

BASE_URL = "https://gdcportalgw.its-mo.com/api_v230317_NE/gdc/"

log = logging.getLogger(__name__)


# from http://stackoverflow.com/questions/17134100/python-blowfish-encryption
def _PKCS5Padding(string):
    byteNum = len(string)
    packingLength = 8 - byteNum % 8
    appendage = chr(packingLength) * packingLength
    return string + appendage


class CarwingsError(Exception):
    pass


class Session(object):
    """Maintains a connection to CARWINGS, refreshing it when needed"""

    def __init__(self, username, password, region="NNA", session: ClientSession=None, base_url=BASE_URL):
        self.username = username
        self.password = password
        self.region_code = region
        self.logged_in = False
        self.custom_sessionid = None
        self.base_url = base_url
        if (session):
            self._shared_session = True
            self.session = session
        else:
            self._shared_session = False
            self.session = ClientSession(
                timeout=ClientTimeout(300, connect=5)
            )

        self._semaphore = Semaphore(1)
        self._connect_timestamp = None

    async def __aenter__(self):
        self.session.headers.update({"User-Agent": ""})
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if not self._shared_session:
            await self.session.close()

    async def _request_with_retry(self, endpoint, params):
        ret = await self._request(endpoint, params)

        if "status" in ret and ret["status"] >= 400:
            current_connect_timestamp = self._connect_timestamp
            async with self._semaphore:  # ensure only one request at a time
                # check the connect timestamp again, as it may have changed while waiting
                if self._connect_timestamp == current_connect_timestamp:
                    log.info("carwings error; logging in and trying request again: %s" % ret)
                    # try logging in again
                    await self.connect()
                ret = await self._request(endpoint, params)

        return ret

    async def _request(self, endpoint, params):
        if self.custom_sessionid:
            params["custom_sessionid"] = self.custom_sessionid
        else:
            params["custom_sessionid"] = ""
            params["initial_app_str"] = "9s5rfKVuMrT03RtzajWNcA"

        url = self.base_url + endpoint

        # Nissan servers can return html instead of jSOn on occassion, e.g.
        #
        # <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//END>
        # <html<head>
        #    <title>503 Service Temporarily Unavailable</title>
        # </head><body>
        # <h1>Service Temporarily Unavailable>
        # <p>The server is temporarily unable to service your
        # request due to maintenance downtime or capacity
        # problems. Please try again later.</p>
        # </body></html>

        try:
            log.debug("invoking carwings API: %s" % url)
            log.debug(
                "params: %s"
                % json.dumps(
                    {
                        k: v.decode("utf-8") if isinstance(v, bytes) else v
                        for k, v in params.items()
                    },
                    sort_keys=True,
                    indent=3,
                    separators=(",", ": "),
                )
            )

            # Nissan servers sometimes do not respond.
            # Connections seem OK, but reads are slow and may not be successful
            async with self.session.post(url, data=params) as response:
                log.debug(
                    "Response HTTP Status Code: {status_code}".format(
                        status_code=response.status
                    )
                )
                log.debug(
                    "Response HTTP Response Body: {content}".format(
                        content=await response.text()
                    )
                )

                try:
                    # ignore the content-type set by the server, as it may be wrong
                    j = await response.json(content_type=None)
                except ContentTypeError as contentTypeError:
                    log.error(
                        "invalid json returned by server: %s" % contentTypeError.message
                    )
                    log.error("response body: %s" % await response.text())
                    raise CarwingsError from contentTypeError

                if "message" in j and j["message"] == "INVALID PARAMS":
                    log.error("carwings error %s: %s" % (j["message"], j["status"]))
                    raise CarwingsError("INVALID PARAMS")

                if "ErrorMessage" in j:
                    log.error(
                        "carwings error %s: %s" % (j["ErrorCode"], j["ErrorMessage"])
                    )
                    raise CarwingsError

                return j
        except ClientError as clientError:
            log.exception(clientError)
            log.warning("HTTP Request failed")
            raise CarwingsError from clientError

    async def connect(self):
        self.custom_sessionid = None
        self.logged_in = False

        response = await self._request(
            "InitialApp_v2.php",
            {
                "RegionCode": self.region_code,
                "lg": "en-US",
            },
        )
        ret = CarwingsInitialAppResponse(response)

        c1 = Blowfish.new(ret.baseprm.encode(), Blowfish.MODE_ECB)
        packedPassword = _PKCS5Padding(self.password)
        encryptedPassword = c1.encrypt(packedPassword.encode())
        encodedPassword = base64.standard_b64encode(encryptedPassword).decode()

        response = await self._request(
            "UserLoginRequest.php",
            {
                "RegionCode": self.region_code,
                "UserId": self.username,
                "Password": encodedPassword,
            },
        )

        ret = CarwingsLoginResponse(response)

        self.custom_sessionid = ret.custom_sessionid

        self.gdc_user_id = ret.gdc_user_id
        log.debug("gdc_user_id: %s" % self.gdc_user_id)
        self.dcm_id = ret.dcm_id
        log.debug("dcm_id: %s" % self.dcm_id)
        self.tz = ret.tz
        log.debug("tz: %s" % self.tz)
        self.language = ret.language
        log.debug("language: %s" % self.language)
        log.debug("vin: %s" % ret.vin)
        log.debug("nickname: %s" % ret.nickname)

        self.leaf = Leaf(self, ret.leafs[0])

        self.logged_in = True
        self._connect_timestamp = datetime.now()

        return ret

    async def get_leaf(self, index=0):
        if not self.logged_in:
            await self.connect()

        return self.leaf


class Leaf:
    def __init__(self, session, params):
        self.session = session
        self.vin = params["vin"]
        self.nickname = params["nickname"]
        self.bound_time = params["bound_time"]
        log.debug("created leaf %s/%s" % (self.vin, self.nickname))

    async def request_update(self):
        response = await self.session._request_with_retry(
            "BatteryStatusCheckRequest.php",
            {
                "RegionCode": self.session.region_code,
                "VIN": self.vin,
            },
        )
        return response["resultKey"]

    async def get_status_from_update(self, result_key):
        response = await self.session._request_with_retry(
            "BatteryStatusCheckResultRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "resultKey": result_key,
            },
        )
        # responseFlag will be "1" if a response has been returned; "0" otherwise
        if response["responseFlag"] == "1":
            return CarwingsBatteryStatusResponse(response)

        return None

    async def start_climate_control(self):
        response = await self.session._request_with_retry(
            "ACRemoteRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
            },
        )
        return response["resultKey"]

    async def get_start_climate_control_result(self, result_key):
        response = await self.session._request_with_retry(
            "ACRemoteResult.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "UserId": self.session.gdc_user_id,  # this userid is the 'gdc' userid
                "resultKey": result_key,
            },
        )
        if response["responseFlag"] == "1":
            return CarwingsStartClimateControlResponse(response)

        return None

    async def stop_climate_control(self):
        response = await self.session._request_with_retry(
            "ACRemoteOffRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
            },
        )
        return response["resultKey"]

    async def get_stop_climate_control_result(self, result_key):
        response = await self.session._request_with_retry(
            "ACRemoteOffResult.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "UserId": self.session.gdc_user_id,  # this userid is the 'gdc' userid
                "resultKey": result_key,
            },
        )
        if response["responseFlag"] == "1":
            return CarwingsStopClimateControlResponse(response)

        return None

    # execute time example: "2016-02-09 17:24"
    # I believe this time is specified in GMT, despite the "tz" parameter
    # TODO: change parameter to python datetime object(?)
    async def schedule_climate_control(self, execute_time):
        response = await self.session._request_with_retry(
            "ACRemoteNewRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "ExecuteTime": execute_time,
            },
        )
        return response["status"] == 200

    # execute time example: "2016-02-09 17:24"
    # I believe this time is specified in GMT, despite the "tz" parameter
    # TODO: change parameter to python datetime object(?)
    async def update_scheduled_climate_control(self, execute_time):
        response = await self.session._request_with_retry(
            "ACRemoteUpdateRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "ExecuteTime": execute_time,
            },
        )
        return response["status"] == 200

    async def cancel_scheduled_climate_control(self):
        response = await self.session._request_with_retry(
            "ACRemoteCancelRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
            },
        )
        return response["status"] == 200

    async def get_climate_control_schedule(self):
        response = await self.session._request_with_retry(
            "GetScheduledACRemoteRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
            },
        )
        if response["status"] == 200:
            if response["ExecuteTime"] != "":
                return CarwingsClimateControlScheduleResponse(response)

        return None

    """
    {
        "status":200,
    }
    """

    async def start_charging(self):
        response = await self.session._request_with_retry(
            "BatteryRemoteChargingRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "ExecuteTime": date.today().isoformat(),
            },
        )
        if response["status"] == 200:
            # This only indicates that the charging command has been received by the
            # Nissan servers, it does not indicate that the car is now charging.
            return True

        return False

    async def get_driving_analysis(self):
        response = await self.session._request_with_retry(
            "DriveAnalysisBasicScreenRequestEx.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
            },
        )
        if response["status"] == 200:
            return CarwingsDrivingAnalysisResponse(response)

        return None

    async def get_latest_battery_status(self):
        response = await self.session._request_with_retry(
            "BatteryStatusRecordsRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "TimeFrom": self.bound_time,
            },
        )
        if response["status"] == 200:
            if "BatteryStatusRecords" in response:
                return CarwingsLatestBatteryStatusResponse(response)
            else:
                log.warning("no battery status record returned by server")
                log.warning("no battery status record returned by server")

        return None

    async def get_latest_hvac_status(self):
        response = await self.session._request_with_retry(
            "RemoteACRecordsRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "TimeFrom": self.bound_time,
            },
        )
        if response["status"] == 200:
            if "RemoteACRecords" in response:
                return CarwingsLatestClimateControlStatusResponse(response)
            else:
                log.warning("no remote a/c records returned by server")
                log.warning("no remote a/c records returned by server")

        return None

    # target_month format: "YYYYMM" e.g. "201602"
    async def get_electric_rate_simulation(self, target_month):
        response = await self.session._request_with_retry(
            "PriceSimulatorDetailInfoRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "TargetMonth": target_month,
            },
        )
        if response["status"] == 200:
            return CarwingsElectricRateSimulationResponse(response)

        return None

    async def request_location(self):
        # As of 25th July the Locate My Vehicle functionality of the Europe version of the
        # Nissan APIs was removed.  It may return, so this call is left here.
        # It currently errors with a 404 MyCarFinderRequest.php was not found on this server
        # for European users.
        response = await self.session._request_with_retry(
            "MyCarFinderRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "UserId": self.session.gdc_user_id,  # this userid is the 'gdc' userid
            },
        )
        return response["resultKey"]

    async def get_status_from_location(self, result_key):
        response = await self.session._request_with_retry(
            "MyCarFinderResultRequest.php",
            {
                "RegionCode": self.session.region_code,
                "lg": self.session.language,
                "DCMID": self.session.dcm_id,
                "VIN": self.vin,
                "tz": self.session.tz,
                "resultKey": result_key,
            },
        )
        if response["responseFlag"] == "1":
            return CarwingsMyCarFinderResponse(response)

        return None



#!/usr/bin/env python

from setuptools import setup, find_packages
import codecs
import os
import re

NAME = 'pycarwings3'

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string")


# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

long_description = read('README.md')

# Note python2 required pycrpto instead of pycryptodome
setup(
    name=NAME,
    version=find_version(NAME, "__init__.py"),
    url='https://github.com/ev-freaks/pycarwings3',
    author='Remus Lazar',
    author_email='remus@ev-freaks.com',
    description='Python library for interacting with the Nissan Leaf Carwings telematics service',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='nissan leaf carwings nissan+you',
    include_package_data=True,
    install_requires=[
        'PyYAML',
        'iso8601',
        'pycryptodome',
        'aiohttp~=3.10.1',
        'pytz',
        'asyncio'],
    packages=find_packages(),
    setup_requires=('pytest-runner', 'ruff'),
    tests_require=['pytest', 'pytest-cov', 'pytest-ruff', 'pytest-asyncio'],
)
